import threading
from _typeshed import Incomplete
from m_crawler.engine.abstract_engine import AbstractEngine as AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse
from m_crawler.pipeline.abstracy_pipeline import AbstractPipeline as AbstractPipeline
from m_crawler.utils.crawler_exception import AllDataCompletedException as AllDataCompletedException, MaxRetryTimeException as MaxRetryTimeException

class Crawler(threading.Thread):
    lock: Incomplete
    engine: Incomplete
    input_pipeline: Incomplete
    output_pipeline: Incomplete
    threads: Incomplete
    max_retry_time: Incomplete
    def __init__(self, engine: AbstractEngine, input_pipeline: AbstractPipeline, output_pipeline: AbstractPipeline, threads: int = 1, name: str = 'Crawler', daemon: bool = True, max_retry_time: int = 5) -> None: ...
    def run(self) -> None: ...



from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse
from typing import Any

class BaseResponse(AbstractResponse):
    def __init__(self, request: AbstractRequest, body: Any, success: bool = True) -> None: ...
    def success_handler(self) -> Any: ...



from abc import ABC, abstractmethod

from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse


class AbstractEngine(ABC):
    """
    引擎,提供与网站实际的交互能力
    """

    def before_download_hook(self, request: AbstractRequest) -> AbstractRequest:
        """
        before_download_hook
        """
        return request

    def after_download_hook(self, result: AbstractResponse) -> AbstractResponse:
        """
        after_download_hook
        """
        return result

    @abstractmethod
    def download(self, request: AbstractRequest) -> AbstractResponse:
        pass

    def run(self, request: AbstractRequest) -> AbstractResponse:
        request = self.before_download_hook(request)
        result = self.download(request)
        result = self.after_download_hook(result)
        return result



from playwright.sync_api import sync_playwright
from m_crawler.config import PlayWeightConfig
from loguru import logger
from m_crawler.engine.abstract_engine import AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.engine.request import PlayWeightRequest
from m_crawler.engine.response import BaseResponse


class PlayWeightEngine(AbstractEngine):
    __slots__ = ("__playwright", "browser", "browser_context", "settings")
    """
    下载引擎
    """

    def __init__(self, settings: PlayWeightConfig) -> None:
        self.settings = settings

    def download(self, request: AbstractRequest) -> AbstractResponse:
        """
        playweight 核心逻辑
        """
        # 确保传入的是 PlayWeightRequest 类型
        if not isinstance(request, PlayWeightRequest):
            raise TypeError("request must be an instance of PlayWeightRequest")
        with sync_playwright() as p:
            browser = p.chromium.launch(
                executable_path=self.settings.executable_path,
                channel=self.settings.channel,
                args=self.settings.args,
                timeout=self.settings.timeout,
                env=self.settings.env,
                headless=self.settings.headless,
                downloads_path=self.settings.downloads_path,
                slow_mo=self.settings.slow_mo,
            )
            context = browser.new_context()
            page = context.new_page()
            try:
                page.goto(request.url)
                for action in request.actions:
                    action(page)
                return BaseResponse(request, page.content(), True)
            except Exception as e:
                logger.error(f"URL:{request.url} failed, error:{e}")
                return BaseResponse(request, None, False)



from typing import Dict, Optional, Union


class PlayWeightConfig:
    __slots__ = (
        "executable_path",
        "channel",
        "args",
        "timeout",
        "env",
        "headless",
        "downloads_path",
        "slow_mo",
    )

    def __init__(self) -> None:
        # 浏览器启动路径
        self.executable_path = r"/opt/google/chrome/google-chrome"
        # chromium, chrome, chrome-beta, chrome-dev...
        self.channel = "chromium"
        # 浏览器启动参数
        self.args = [
            "--disable-blink-features=AutomationControlled",
            "--user-agent=Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/139.0.0.0 Safari/537.36",
        ]
        # 浏览器启动超时时间，单位 ms
        self.timeout = 10000
        # 环境变量
        self.env: Optional[Dict[str, Union[str, float, bool]]] = None
        # headless 模式
        self.headless = True
        # download 下载地址，不指定当触发下载时将现在到临时目录，且在退出时删除
        self.downloads_path = ""
        # 运行速度减慢指定的毫秒数
        self.slow_mo = 0



import threading


class CountDown:
    __slots__ = "__count"
    __lock = threading.Lock()

    def __init__(self, count: int = 0) -> None:
        self.__count = count

    def count_down(self) -> bool:
        with CountDown.__lock:
            if self.__count <= 0:
                return True
            return False

    def add(self) -> None:
        with CountDown.__lock:
            self.__count += 1

    def down(self) -> None:
        with CountDown.__lock:
            self.__count -= 1



from concurrent.futures import Future, ThreadPoolExecutor, as_completed
import threading
from loguru import logger

from m_crawler.engine.abstract_engine import AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.pipeline.abstracy_pipeline import AbstractPipeline
from m_crawler.utils.crawler_exception import (
    AllDataCompletedException,
    MaxRetryTimeException,
)


class Crawler(threading.Thread):

    lock = threading.Lock()

    def __init__(
        self,
        engine: AbstractEngine,
        input_pipeline: AbstractPipeline,
        output_pipeline: AbstractPipeline,
        threads: int = 1,
        name: str = "Crawler",
        daemon: bool = True,
        max_retry_time: int = 5,
    ) -> None:
        super().__init__(name=name, daemon=daemon)
        self.engine = engine
        self.input_pipeline = input_pipeline
        self.output_pipeline = output_pipeline
        self.threads = threads
        self.max_retry_time = max_retry_time
        self.__retry_dict: dict[str, int] = {}

    def run(self) -> None:
        futures: list[Future[None]] = []
        with ThreadPoolExecutor(max_workers=self.threads) as executor:
            while True:
                resuest = self.input_pipeline.queue.get(block=True)
                if resuest is None:
                    break
                futures.append(executor.submit(self.__run, resuest))

        for future in as_completed(futures):
            future.result()
        self.output_pipeline.proxy_process(AllDataCompletedException())

    def __run(self, request: AbstractRequest) -> None:
        response: AbstractResponse = None  # type: ignore
        try:
            response = self.engine.run(request)
            if not response.success:
                raise RuntimeError()
        except Exception as e:
            with Crawler.lock:
                logger.warning(f"{request.url} error: {e}")
                self.__retry_dict[request.url] = (
                    self.__retry_dict.get(request.url, 0) + 1
                )
                if self.__retry_dict[request.url] > self.max_retry_time:
                    logger.error(f"Max retry time reached for {request.url}")
                    raise MaxRetryTimeException()
                logger.warning(
                    f"Retrying {request.url}, Current retry count: {self.__retry_dict[request.url]}"
                )
            self.__run(request)
        self.output_pipeline.proxy_process(response)



from _typeshed import Incomplete
from m_crawler.config import PlayWeightConfig as PlayWeightConfig
from m_crawler.engine.abstract_engine import AbstractEngine as AbstractEngine
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse
from m_crawler.engine.request import PlayWeightRequest as PlayWeightRequest
from m_crawler.engine.response import BaseResponse as BaseResponse

class PlayWeightEngine(AbstractEngine):
    settings: Incomplete
    def __init__(self, settings: PlayWeightConfig) -> None: ...
    def download(self, request: AbstractRequest) -> AbstractResponse: ...



import abc
from abc import ABC, abstractmethod
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse

class AbstractEngine(ABC, metaclass=abc.ABCMeta):
    def before_download_hook(self, request: AbstractRequest) -> AbstractRequest: ...
    def after_download_hook(self, result: AbstractResponse) -> AbstractResponse: ...
    @abstractmethod
    def download(self, request: AbstractRequest) -> AbstractResponse: ...
    def run(self, request: AbstractRequest) -> AbstractResponse: ...



from abc import ABC
from typing import Any, Optional


class AbstractRequest(ABC):
    """
    AbstractRequest
    url: str 请求的 URL
    headers: dict[str, str] 请求头
    params: dict[str, Any] 请求参数
    meta: dict[str, Any] 请求的元数据，通常用于爬虫的数据传递
    """

    __slots__ = ("url", "headers", "params", "meta")

    def __init__(
        self,
        url: str,
        headers: Optional[dict[str, str]] = None,
        params: Optional[dict[str, Any]] = None,
        meta: Optional[dict[str, Any]] = None,
    ) -> None:
        self.url = url
        if headers:
            self.headers = headers
        if params:
            self.params = params
        if meta:
            self.meta = meta



from _typeshed import Incomplete
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from playwright.sync_api import Page as Page
from typing import Any, Callable, Sequence

class PlayWeightRequest(AbstractRequest):
    download: Incomplete
    actions: Sequence[Callable[[Page], Any]]
    def __init__(self, url: str, headers: dict[str, str] | None = None, params: dict[str, Any] | None = None, download: bool = False, actions: Sequence[Callable[[Page], Any]] = []) -> None: ...



from _typeshed import Incomplete

class PlayWeightConfig:
    executable_path: str
    channel: str
    args: Incomplete
    timeout: int
    env: dict[str, str | float | bool] | None
    headless: bool
    downloads_path: str
    slow_mo: int
    def __init__(self) -> None: ...



class MaxRetryTimeException(Exception):
    def __init__(self, message: str = 'Max retry count reached') -> None: ...

class AllDataCompletedException(Exception):
    def __init__(self, message: str = 'All data completed') -> None: ...



from abc import ABC, abstractmethod
from typing import Any
from loguru import logger

from m_crawler.engine.abstract_request import AbstractRequest


class AbstractResponse(ABC):
    def __init__(
        self, request: AbstractRequest, body: Any, success: bool = True
    ) -> None:
        self.request = request
        self.body = body
        self.success = success

    @abstractmethod
    def success_handler(self) -> Any:
        pass

    def fail_handler(self, exception: Exception) -> None:
        logger.error(f"{self.request.url} fetch error, {exception}")

    def handle(self) -> Any:
        try:
            return self.success_handler()
        except Exception as e:
            self.fail_handler(e)



from typing import Any, Callable, Optional, Sequence
from playwright.sync_api import Page

from m_crawler.engine.abstract_request import AbstractRequest


class PlayWeightRequest(AbstractRequest):
    __slots__ = ("url", "headers", "params", "download", "actions")

    def __init__(
        self,
        url: str,
        headers: Optional[dict[str, str]] = None,
        params: Optional[dict[str, Any]] = None,
        download: bool = False,
        actions: Sequence[Callable[[Page], Any]] = [],
    ) -> None:
        super().__init__(url, headers, params)
        self.download = download
        self.actions: Sequence[Callable[[Page], Any]] = actions



from abc import ABC, abstractmethod
from queue import Queue
from typing import Iterable, Optional, Union

from m_crawler.engine.abstract_request import AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.utils.crawler_exception import AllDataCompletedException


class AbstractPipeline(ABC):

    def __init__(self) -> None:
        self.queue: Queue[Optional[AbstractRequest]] = Queue()

    def proxy_process(
        self, item: Union[AbstractResponse, AllDataCompletedException]
    ) -> None:
        if isinstance(item, AllDataCompletedException):
            self.queue.put(None)
        else:
            process_result = self.process(item)
            if process_result:
                for request in process_result:
                    self.queue.put(request)

    @abstractmethod
    def process(self, item: AbstractResponse) -> Optional[Iterable[AbstractRequest]]:
        """
        处理数据
        """
        pass



import abc
from abc import ABC, abstractmethod
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from m_crawler.engine.abstract_response import AbstractResponse as AbstractResponse
from m_crawler.utils.crawler_exception import AllDataCompletedException as AllDataCompletedException
from queue import Queue
from typing import Iterable

class AbstractPipeline(ABC, metaclass=abc.ABCMeta):
    queue: Queue[AbstractRequest | None]
    def __init__(self) -> None: ...
    def proxy_process(self, item: AbstractResponse | AllDataCompletedException) -> None: ...
    @abstractmethod
    def process(self, item: AbstractResponse) -> Iterable[AbstractRequest] | None: ...



from _typeshed import Incomplete
from abc import ABC
from typing import Any

class AbstractRequest(ABC):
    url: Incomplete
    headers: Incomplete
    params: Incomplete
    meta: Incomplete
    def __init__(self, url: str, headers: dict[str, str] | None = None, params: dict[str, Any] | None = None, meta: dict[str, Any] | None = None) -> None: ...



class MaxRetryTimeException(Exception):
    def __init__(self, message: str = "Max retry count reached") -> None:
        super().__init__(message)

class AllDataCompletedException(Exception):
    def __init__(self, message: str = "All data completed") -> None:
        super().__init__(message)



from m_crawler.engine.abstract_response import AbstractResponse
from m_crawler.engine.abstract_request import AbstractRequest
from typing import Any


class BaseResponse(AbstractResponse):
    def __init__(
        self, request: AbstractRequest, body: Any, success: bool = True
    ) -> None:
        super().__init__(request, body, success)

    def success_handler(self) -> Any:
        return self.body



import abc
from _typeshed import Incomplete
from abc import ABC, abstractmethod
from m_crawler.engine.abstract_request import AbstractRequest as AbstractRequest
from typing import Any

class AbstractResponse(ABC, metaclass=abc.ABCMeta):
    request: Incomplete
    body: Incomplete
    success: Incomplete
    def __init__(self, request: AbstractRequest, body: Any, success: bool = True) -> None: ...
    @abstractmethod
    def success_handler(self) -> Any: ...
    def fail_handler(self, exception: Exception) -> None: ...
    def handle(self) -> Any: ...
